/// @addtogroup diffIncl
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file DiffInclusion.h
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

/* Author: Tomasz Kapela, 2007 */

#ifndef _CAPD_DIFFINCL_DIFFINCLUSION_H_ 
#define _CAPD_DIFFINCL_DIFFINCLUSION_H_ 

#include <string>
#include <map>
#include "capd/vectalg/Norm.h"
#include "capd/dynsys/Taylor.h"
#include "capd/dynsys/enclosure.h"

namespace capd{
  namespace diffIncl{
    
    template<typename MapT>
    class DiffInclusion  {
      
    public:
      typedef MapT MultiMapType;
      typedef MultiMapType MapType;
      typedef typename MapT::MapType VectorFieldType;
      typedef typename MapT::FunctionType FunctionType;
      typedef typename MapType::MatrixType MatrixType;
      typedef typename MatrixType::RowVectorType VectorType;
      typedef typename MatrixType::ScalarType ScalarType;
      typedef capd::vectalg::Norm<VectorType, MatrixType> NormType;
      typedef capd::dynsys::Taylor<VectorFieldType> DynamicalSystemType;
      
      DiffInclusion(   MultiMapType& diffIncl,               // map of the form f(x)+ g(x,e)
              int Order,                            // order for integration
              ScalarType const & Step,              // time step for integration
              NormType const & norm                 // norm
              );
      
      ~DiffInclusion();
      
      VectorType diffInclusionEnclosure(const VectorType& x) const;
      VectorType dynamicalSystemEnclosure(const VectorType& x) const;
      
      VectorType enclosure(const VectorType & x) const;
      
      // get & set functions for vector field
      const MapType& getField() const;
      MapType& getField();
      
      // get & set functions for Taylor method order
      int getOrder() const;
      void setOrder(int newOrder);
      
      // get & set functions for time step
      const ScalarType& getStep() const;
      void setStep(const ScalarType& newStep);
      
      DynamicalSystemType & getDynamicalSystem();
      const DynamicalSystemType & getDynamicalSystem() const;
      
    protected:
    //  void operator=(const DiffInclusion & a_t){}
    // DiffInclusion(const DiffInclusion & diffIncl) : m_norm(m_norm.clone()), m_diffIncl(diffIncl.m_diffIncl){}
      NormType * m_norm;
      MultiMapType & m_diffIncl;
      DynamicalSystemType m_dynamicalSystem;
    };
    
    
// --------------- inline definitions -----------------
    
    
    template<typename MapType>
    inline const MapType& DiffInclusion<MapType>::getField() const {
      return m_diffIncl;
    }
    
    template<typename MapType>
    inline MapType & DiffInclusion<MapType>::getField() {
      return m_diffIncl;
    }
    
   
    template<typename MapType>
    inline int DiffInclusion<MapType>::getOrder() const {
      return m_dynamicalSystem.getOrder();
    }
    
    template<typename MapType>
    inline void DiffInclusion<MapType>::setOrder(int newOrder) {
      m_dynamicalSystem.setOrder(newOrder);
    }
    
    
    template<typename MapType>
    inline const typename DiffInclusion<MapType>::ScalarType& DiffInclusion<MapType>::getStep() const {
      return m_dynamicalSystem.getStep();
    }
    
    template<typename MapType>
    inline void DiffInclusion<MapType>::setStep(const ScalarType& newStep) {
      m_dynamicalSystem.setStep(newStep);
    }
    
/** 
 * Computes enclosure of image of given set for differential inclusion during whole time step
 */
    template<typename MapT>
    inline typename DiffInclusion<MapT>::VectorType DiffInclusion<MapT>::diffInclusionEnclosure(const VectorType& x) const{
      
      return capd::dynsys::enclosure(m_diffIncl, x, getStep());
    }
    
    template <typename MapT>
    inline typename DiffInclusion<MapT>::VectorType DiffInclusion<MapT>::dynamicalSystemEnclosure(const VectorType & x) const {
      return m_dynamicalSystem.enclosure(x);
    }

    template<typename MapT>
    inline typename DiffInclusion<MapT>::VectorType DiffInclusion<MapT>::enclosure(const VectorType& x) const{
      
      return diffInclusionEnclosure(x);
    }
    
    template <typename MapT>
    inline typename DiffInclusion<MapT>::DynamicalSystemType & DiffInclusion<MapT>::getDynamicalSystem(){
      return m_dynamicalSystem;
    }
    
    template <typename MapT>
    inline const typename DiffInclusion<MapT>::DynamicalSystemType&  DiffInclusion<MapT>::getDynamicalSystem() const{
      return m_dynamicalSystem;
    }
    
    
//###########################################################//
    
    
  }} // namespace capd::diffIncl

#endif // _CAPD_DIFFINCL_DIFFINCLUSION_H_ 

/// @}
