/// @addtogroup diffIncl
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file InclRect2Set.h
///
/// @author Tomasz Kapela
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2007 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#ifndef _CAPD_DIFFINCL_INCLRECT2SET_H_ 
#define _CAPD_DIFFINCL_INCLRECT2SET_H_ 

//#include "dynset/inclc0set.h"
#include "capd/dynset/Rect2Set.h"
#include "capd/vectalg/Norm.h"

namespace capd{
namespace diffIncl{

///////////////////////////////////////////////////////////////////////////
// InclRect2Set
/// 
/// Set representation for differential inclusions based on capd::dynset::Rect2Set class
///
///   set is represented as: x + C*r0 + B*r   where
///       C*r0 - basic 'Lipschitz part'
///       B*r - QR-decomposition of the remaining errors
///   

template<typename MatrixT>
class InclRect2Set : public capd::dynset::Rect2Set<MatrixT>{

public:
  typedef MatrixT MatrixType;
  typedef typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;
  typedef capd::vectalg::Norm<VectorType,MatrixType> NormType;

  // constr
  explicit InclRect2Set(int, double s_factor = 0.);
  explicit InclRect2Set(const VectorType& the_x, double s_factor = 0.);
  InclRect2Set(const VectorType& the_x, const VectorType& the_r0, double s_factor = 0.);
  InclRect2Set(const VectorType& the_x, const MatrixType& the_C, const VectorType& the_r0, double s_factor = 0.);
  InclRect2Set(const VectorType& the_x, const MatrixType& the_C,
               const VectorType& the_r0,
               const VectorType& the_r, double s_factor = 0.
  );
  
  capd::dynset::C0Set<MatrixType>* clone(void);
  capd::dynset::C0Set<MatrixType>* fresh(void);
  capd::dynset::C0Set<MatrixType>* cover(const VectorType& v);
  
  template<typename DiffIncl>
  void move( DiffIncl& dynsys);
  template<typename DiffIncl>
  void move( DiffIncl & dynsys, InclRect2Set& result) const;
  
  std::vector<VectorType> getCorners() const;

  using capd::dynset::Rect2Set<MatrixT>::get_x;
  using capd::dynset::Rect2Set<MatrixT>::get_r;
  using capd::dynset::Rect2Set<MatrixT>::get_r0;
  using capd::dynset::Rect2Set<MatrixT>::get_B;
  using capd::dynset::Rect2Set<MatrixT>::get_C;
  using capd::dynset::Rect2Set<MatrixT>::size;
  using capd::dynset::Rect2Set<MatrixT>::center;
  using capd::dynset::Rect2Set<MatrixT>::operator VectorType;
  using capd::dynset::Rect2Set<MatrixT>::show;
  using capd::dynset::Rect2Set<MatrixT>::affineTransformation;

protected:
  using capd::dynset::Rect2Set<MatrixT>::m_x;
  using capd::dynset::Rect2Set<MatrixT>::m_r;
  using capd::dynset::Rect2Set<MatrixT>::m_r0;
  using capd::dynset::Rect2Set<MatrixT>::m_B;
  using capd::dynset::Rect2Set<MatrixT>::m_C;
};

template<typename MatrixType>
std::vector<typename MatrixType::VectorType> getCorners(const InclRect2Set<MatrixType> & set) ;

// inline definitions
////////////////////////////////////////////////////////////////////////////////
/// Constructors
template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(int dim, double s_factor)
  :  capd::dynset::Rect2Set<MatrixType>(dim, s_factor) {
}

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(const VectorType& the_x, double s_factor)
  :  capd::dynset::Rect2Set<MatrixType>(the_x, s_factor) {
}

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(const VectorType& the_x,const VectorType& the_r0, double s_factor)
  :  capd::dynset::Rect2Set<MatrixType>(the_x, the_r0, s_factor) {
}

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(
      const VectorType& the_x,
      const MatrixType& the_C,
      const VectorType& the_r0, 
      double s_factor
   )
  : capd::dynset::Rect2Set<MatrixType>(the_x, the_C, the_r0, s_factor){
}

template<typename MatrixType>
inline InclRect2Set<MatrixType>::InclRect2Set(
      const VectorType& the_x,
      const MatrixType &the_C,
      const VectorType& the_r0,
      const VectorType& the_r,
      double s_factor
   ): capd::dynset::Rect2Set<MatrixType>(the_x, the_C, the_r0, the_r, s_factor){
}

////////////////////////////////////////////////////////////////////////////////
/// C0set interface overriding
   
template<typename MatrixType>
inline typename capd::dynset::C0Set<MatrixType>* InclRect2Set<MatrixType>::clone(){
   return new InclRect2Set(*this);
}

template<typename MatrixType>
inline typename capd::dynset::C0Set<MatrixType>* InclRect2Set<MatrixType>::fresh(){
   return new InclRect2Set<MatrixType>(m_x.dimension());
}

template<typename MatrixType>
inline typename capd::dynset::C0Set<MatrixType>* InclRect2Set<MatrixType>::cover(const VectorType& v){
   return new InclRect2Set(midVector(v),v-midVector(v));
}
////////////////////////////////////////////////////////////////////////////////
template<typename MatrixType>
std::vector<typename MatrixType::VectorType> getCorners(const capd::diffIncl::InclRect2Set<MatrixType> & set) ;

}} // namespace capd::diffIncl

#endif // _CAPD_DIFFINCL_INCLRECT2SET_H_ 

/// @}
