/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file C1Rect.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

/* Author: Daniel Wilczak 2002-2004 */

#ifndef _CAPD_DYNSET_C1RECT_H_ 
#define _CAPD_DYNSET_C1RECT_H_ 

#include "capd/dynset/C1Set.h"
#include "capd/vectalg/Norm.h"
#include "capd/dynsys/C1DynSys.h"


/* 
  C^1-Lohner algorithm.
  Derivative of the flow moved via QR decomposition (third method)
  the set part - QR decomposition (3-rd method)
*/

namespace capd{
namespace dynset{

template<typename MatrixT>
class C1Rect : public C1Set<MatrixT>{
public:
  typedef MatrixT MatrixType;
  typedef typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;
  typedef capd::vectalg::Norm<VectorType,MatrixType> NormType;

  C1Rect(const NormType& aNorm, int dim);
  C1Rect(const VectorType& x, const NormType& aNorm);
  C1Rect(const VectorType& x, const VectorType& r, const NormType& aNorm);
  C1Rect(const VectorType& x, const MatrixType& B, const VectorType& r, const NormType& aNorm);
  C1Rect(const C1Rect& c);
  ~C1Rect(){delete m_Norm;}

  ScalarType size(void) const;
  C1Set<MatrixType>* clone(void) const;
  C1Set<MatrixType>* fresh(void) const;
  C1Set<MatrixType>* cover(const VectorType& v) const;
  VectorType center(void) const;
  void move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys);
  void move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys, C1Rect& result) const;

  std::string show(void) const;
  const NormType *getNorm(void) const;
  operator VectorType() const;
  operator MatrixType() const;
  C1Rect& operator=(const VectorType &v);
  C1Rect& operator=(const C1Rect &S);

protected:

// representation of the set: x + B * r

  VectorType m_x,m_r;
  MatrixType m_B;

  MatrixType m_D, m_Bjac, m_R; // D + Bjac * R

  NormType *m_Norm; //the logarithmic norm used in the computation of a rough
         // enclosure for the variational part
};

// ------------------------------------------------

template<typename MatrixType>
inline void C1Rect<MatrixType>::move(capd::dynsys::C1DynSys<MatrixType> &c1dynsys)
{
  move(c1dynsys,*this);
}

template<typename MatrixType>
inline C1Set<MatrixType> * C1Rect<MatrixType>::clone(void) const
{
  return new C1Rect<MatrixType>(*this);
}

template<typename MatrixType>
inline C1Set<MatrixType> *C1Rect<MatrixType>::fresh(void) const
{
  return new C1Rect<MatrixType>(*m_Norm,m_x.dimension());
}

template<typename MatrixType>
inline C1Set<MatrixType> *C1Rect<MatrixType>::cover(const VectorType &v) const
{
  return new C1Rect(midVector(v),v-midVector(v), *m_Norm);
}

template<typename MatrixType>
inline typename C1Rect<MatrixType>::VectorType C1Rect<MatrixType>::center(void) const
{
  return m_x;
}

template<typename MatrixT>
inline C1Rect<MatrixT>::operator typename C1Rect<MatrixT>::VectorType(void) const
{
  return m_x + m_B*m_r;
}

template<typename MatrixType>
inline const typename C1Rect<MatrixType>::NormType *C1Rect<MatrixType>::getNorm(void) const
{
  return m_Norm;
}

}} // the end of the namespace capd::dynset

#endif // _CAPD_DYNSET_C1RECT_H_ 

/// @}
