/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file C1Rect2.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

/* Author: Daniel Wilczak 2002-2004 */

#ifndef _CAPD_DYNSET_C1RECT2_H_ 
#define _CAPD_DYNSET_C1RECT2_H_ 

#include "capd/dynset/C1Set.h"
#include "capd/vectalg/Norm.h"
#include "capd/dynsys/C1DynSys.h"

/* C^1-Lohner algorithm.
   Derivative of the flow moved via QR decomposition (third method)
   the set part - QR decomposition (3-rd method)
*/

namespace capd{
namespace dynset{

template<typename MatrixT>
class C1Rect2 : public C1Set<MatrixT>{
public:
  typedef MatrixT MatrixType;
  typedef typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;
  typedef capd::vectalg::Norm<VectorType,MatrixType> NormType;

  C1Rect2(const NormType& the_norm, int);
  C1Rect2(const VectorType& the_x, const NormType&, double s_factor = 20);
  C1Rect2(const VectorType& the_x, const VectorType& the_r, const NormType&, double s_factor = 20);
  C1Rect2(const VectorType& the_x, const MatrixType&, const VectorType& the_r, const NormType&, double s_factor = 20);
  C1Rect2(const C1Rect2 &c);
  ~C1Rect2(){delete m_Norm;}

  ScalarType size(void) const;
  C1Set<MatrixType>* clone(void) const;
  C1Set<MatrixType>* fresh(void) const;
  C1Set<MatrixType>* cover(const VectorType& v) const;

  VectorType center(void) const;
  void move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys);
  void move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys, C1Rect2& result) const;

  std::string show(void) const;
  const NormType* getNorm(void) const;
  double setFactor(double sFactor); // sets size factor :  0 = no swapping
  void disableSwapping();
  double getFactor();

  operator VectorType() const;
  operator MatrixType() const;
  C1Rect2 &operator=(const VectorType &v);
  C1Rect2 &operator=(const C1Rect2 &S);
  
protected:

// a representation of the set: x + C*r0 + B*r

  VectorType m_x,m_r,m_r0;
  MatrixType m_B,m_C;
  VectorType m_currentSet;
  MatrixType m_currentMatrix;
  
  MatrixType m_D, m_Bjac, m_R, m_Cjac, m_R0; // D + Bjac*R  + Cjac * R0;
  
  NormType *m_Norm; //the logarithmic Norm<IVectorType,IMatrixType> used in the computation of a rough
                  // enclosure for the variational part
  
  double m_sizeFactor;  // if size of r0 is greater than r *size_factor we  put r in r0 and set r=0
                     // if size_factor = 0 we disable swapping
};

// -------------------------------------------------------------------

template<typename MatrixType>
inline void C1Rect2<MatrixType>::move(capd::dynsys::C1DynSys<MatrixType>& c1dynsys)
{
  move(c1dynsys,*this);
}

template<typename MatrixType>
inline C1Set<MatrixType>* C1Rect2<MatrixType>::clone(void) const
{
  return new C1Rect2<MatrixType>(*this);
}

template<typename MatrixType>
inline C1Set<MatrixType>* C1Rect2<MatrixType>::fresh(void) const
{
  return new C1Rect2<MatrixType>(*m_Norm,m_x.dimension());
}

template<typename MatrixType>
inline C1Set<MatrixType>* C1Rect2<MatrixType>::cover(const VectorType& v) const
{
  return new C1Rect2<MatrixType>(v, *m_Norm);
}

template<typename MatrixType>
inline typename C1Rect2<MatrixType>::VectorType C1Rect2<MatrixType>::center(void) const
{
  return m_x;
}

template<typename MatrixType>
inline C1Rect2<MatrixType>::operator typename C1Rect2<MatrixType>::VectorType(void) const
{
  return m_currentSet;
}

template<typename MatrixType>
inline const typename C1Rect2<MatrixType>::NormType* C1Rect2<MatrixType>::getNorm(void) const
{
  return m_Norm;
}

template<typename MatrixType>
inline double C1Rect2<MatrixType>::setFactor(double sFactor) // sets size factor :  0 = no swapping
{
  double s = m_sizeFactor;
  m_sizeFactor = sFactor;
  return s;
}

template<typename MatrixType>
inline void C1Rect2<MatrixType>::disableSwapping()
{
  m_sizeFactor = 0.0;
}

template<typename MatrixType>
inline double C1Rect2<MatrixType>::getFactor() // returns size factor
{
  return m_sizeFactor;
}

}} // the end of the namespace capd::dynset

#endif // _CAPD_DYNSET_C1RECT2_H_ 

/// @}
