/// @addtogroup dynset
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file C1Rect2.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

#ifndef _CAPD_DYNSET_C1RECT2_HPP_ 
#define _CAPD_DYNSET_C1RECT2_HPP_ 

#include <sstream>
#include "capd/capd/minmax.h"
#include "capd/dynset/C1Rect2.h"
#include "capd/matrixAlgorithms/floatMatrixAlgorithms.hpp"

namespace capd{
namespace dynset{

template<typename MatrixType>
C1Rect2<MatrixType>::C1Rect2(const NormType& aNorm, int dim)
  : m_x(dim), m_r(dim), m_r0(dim),
    m_B(MatrixType::Identity(dim)),
    m_C(MatrixType::Identity(dim)),
    m_currentSet(dim),
    m_currentMatrix(dim,dim),
    m_D(MatrixType::Identity(dim)),
    m_Bjac(MatrixType::Identity(dim)),
    m_R(dim,dim),
    m_Cjac(MatrixType::Identity(dim)),
    m_R0(dim,dim),
    m_Norm(aNorm.clone()),
    m_sizeFactor(20.)
{
  m_x.clear();
  m_r.clear();
  m_r0.clear();
  m_R.clear();
  m_R0.clear();
  m_currentSet.clear();
  m_currentMatrix.clear();
}


template<typename MatrixType>
C1Rect2<MatrixType>::C1Rect2(const C1Rect2& s)
  : m_x(s.m_x),
    m_r(s.m_r),
    m_r0(s.m_r0),
    m_B(s.m_B),
    m_C(s.m_C),
    m_currentSet(s.m_currentSet),
    m_currentMatrix(s.m_currentMatrix),
    m_D(s.m_D),
    m_Bjac(s.m_Bjac),
    m_R(s.m_R),
    m_Cjac(s.m_Cjac),
    m_R0(s.m_R0),
    m_Norm(s.m_Norm->clone()),
    m_sizeFactor(s.m_sizeFactor)
{}

template<typename MatrixType>
C1Rect2<MatrixType>::C1Rect2(const VectorType& x, const NormType& aNorm, double sizeFactor)
 : m_x(x),
   m_r(x.dimension()),
   m_r0(x.dimension()),
   m_B(MatrixType::Identity(x.dimension())),
   m_C(MatrixType::Identity(x.dimension())),
   m_currentSet(x.dimension()),
   m_currentMatrix(x.dimension(),x.dimension()),
   m_D(MatrixType::Identity(x.dimension())),
   m_Bjac(MatrixType::Identity(x.dimension())),
   m_R(x.dimension(),x.dimension()),
   m_Cjac(MatrixType::Identity(x.dimension())),
   m_R0(x.dimension(),x.dimension()),
   m_Norm(aNorm.clone()),
   m_sizeFactor(sizeFactor)
{
  m_currentSet = m_x;
  split(m_x,m_r0);
  m_r.clear();
  m_R.clear();
  m_R0.clear();
  m_currentMatrix.setToIdentity();
}

template<typename MatrixType>
C1Rect2<MatrixType>::C1Rect2(
  const VectorType& x,
  const VectorType& r0,
  const NormType& aNorm,
  double sizeFactor
  ):m_x(x),
    m_r(x.dimension()),
    m_r0(r0),
    m_B(MatrixType::Identity(x.dimension())),
    m_C(MatrixType::Identity(x.dimension())),
    m_currentMatrix(MatrixType::Identity(x.dimension())),
    m_D(MatrixType::Identity(x.dimension())),
    m_Bjac(MatrixType::Identity(x.dimension())),
    m_R(x.dimension(),x.dimension()),
    m_Cjac(MatrixType::Identity(x.dimension())),
    m_R0(x.dimension(),x.dimension()),
    m_Norm(aNorm.clone()),
    m_sizeFactor(sizeFactor)
{
  m_x += m_r0;
  m_currentSet = m_x;
  split(m_x,m_r0);
  m_r.clear();
  m_R.clear();
  m_R0.clear();  
}


template<typename MatrixType>
C1Rect2<MatrixType>::C1Rect2(
  const VectorType& x,
  const MatrixType& C,
  const VectorType& r0,
  const NormType& aNorm,
  double sizeFactor
  ):m_x(x),
    m_r(x.dimension()),
    m_r0(r0),
    m_B(MatrixType::Identity(x.dimension())),
    m_C(C),
    m_currentMatrix(MatrixType::Identity(x.dimension())),
    m_D(MatrixType::Identity(x.dimension())),
    m_Bjac(MatrixType::Identity(x.dimension())),
    m_R(x.dimension(),x.dimension()),
    m_Cjac(MatrixType::Identity(x.dimension())),
    m_R0(x.dimension(),x.dimension()),
    m_Norm(aNorm.clone()),
    m_sizeFactor(sizeFactor)
{
  m_r.clear();
  m_R.clear();
  m_R0.clear();
// we must assure that m_x\in m_x + m_C*m_r0
  m_currentSet = m_x + m_C*m_r0;
  if(!subset(m_r,m_r0))
  {
    VectorType centerR0=m_r0;
    split(centerR0,m_r0);
    // now the m_r0 is centered. We must add C*centerR0 to m_x
    m_x += m_C*centerR0;
    // finally, we split m_x to the point ane the rest is added to r
    capd::vectalg::split(m_x,m_r);
  }
}

//-----------------------------------------------

template<typename MatrixType>
void C1Rect2<MatrixType>::move(capd::dynsys::C1DynSys<MatrixType> &c1dynsys, C1Rect2 &result) const
{
  int dim = m_x.dimension();
  VectorType Rem(dim);
  MatrixType jacRem(dim,dim);

  VectorType xx = m_currentSet;
  
  int found;
  VectorType Enc = c1dynsys.enclosure(xx,&found);
  MatrixType S = c1dynsys.jacEnclosure(Enc,*m_Norm);
  c1dynsys.JacRemainder(Enc,S,Rem,jacRem);
  MatrixType Jac = c1dynsys.JacPhi(xx);

  result.m_x = c1dynsys.Phi(m_x) + Rem;
  result.m_C = Jac*m_C;
  MatrixType A = Jac*m_B;
  result.m_currentSet = result.m_x + result.m_C*m_r0 + A*m_r;

// xx is unnecesarry now
  capd::vectalg::split(result.m_x,xx);
  capd::vectalg::split(result.m_C,S);
  xx += S*m_r0;
// S is unnecessary now
  capd::vectalg::mid(A,result.m_B); // result.m_B = midMatrix(A) but faster
  capd::matrixAlgorithms::orthonormalize(result.m_B);
  MatrixType Qtr = Transpose(result.m_B);
  result.m_r = (Qtr*A)*m_r + Qtr*xx;

  if(&result!=this) result.m_r0 = m_r0;
  // T.Kapela - put set r into r0 if r is too big.
  if(m_sizeFactor!=0)
    if( capd::vectalg::size(result.m_r) > m_sizeFactor * capd::vectalg::size(result.m_r0))
    {
      result.m_r0 = result.m_r +  ((Qtr * result.m_C) * result.m_r0);
      result.m_C = result.m_B;
      result.m_B = MatrixType::Identity(dim);
      result.m_r = ScalarType(0.);
    }

//-----------------------------------------------

  Jac += jacRem;
  result.m_D = Jac*m_D;
  result.m_Cjac = Jac * m_Cjac;
  MatrixType X = Jac*m_Bjac;
  result.m_currentMatrix = result.m_D + result.m_Cjac*m_R0 + X*m_R;
  split(result.m_Cjac,S);
  result.m_D +=  S*m_R0;

  split(result.m_D,S);

  // S is unnecessary now, so we will use it again
  mid(X,result.m_Bjac);
  capd::matrixAlgorithms::orthonormalize(result.m_Bjac);
  Qtr = Transpose(result.m_Bjac);
  result.m_R = (Qtr*X)*m_R + Qtr*S;

  if (&result!=this) 
    result.m_R0 = m_R0;

  if(m_sizeFactor!=0)
  {
    typename ScalarType::BoundType maxSizeR = 0., maxSizeR0 = 0.;
    int i,j;
    for(i=0;i<dim;i++)
      for(j=0;j<dim;j++)
      {
        maxSizeR = capd::max(maxSizeR,diam(result.m_R[i][j]).rightBound());
        maxSizeR0 = capd::max(maxSizeR0, diam(result.m_R0[i][j]).rightBound());
      }

    if(maxSizeR>m_sizeFactor*maxSizeR0)
    {
      result.m_R0 = result.m_R + (Qtr * result.m_Cjac) * result.m_R0;
      result.m_Cjac = result.m_Bjac;
      result.m_Bjac = MatrixType::Identity(dim);
      result.m_R.clear();
    }
  } 
}

// ------------------------------------------------------------------------

template<typename MatrixType>
typename C1Rect2<MatrixType>::ScalarType C1Rect2<MatrixType>::size(void) const
{
  return capd::vectalg::size(m_currentSet);
}

template<typename MatrixType>
std::string C1Rect2<MatrixType>::show(void) const
{
  std::ostringstream descriptor;
  descriptor << "C1Rect2:\n x=";
  descriptor << m_x << "\n B=";
  descriptor << m_B << "\n r=";
  descriptor << m_r << "\n C= ";
  descriptor << m_C << "\n r0= ";
  descriptor << m_r0 << "\n D= ";
  descriptor << m_D << "\n Bjac= ";
  descriptor << m_Bjac << "\n R= ";
  descriptor << m_R << "\n Cjac= ";
  descriptor << m_Cjac << "\n R0= ";
  descriptor << m_R0 << '\0';
  
  return descriptor.str();
}

template<typename MatrixT>
C1Rect2<MatrixT>::operator typename C1Rect2<MatrixT>::MatrixType() const
{
  return m_currentMatrix;
}

template<typename MatrixType>
C1Rect2<MatrixType> &C1Rect2<MatrixType>::operator=(const VectorType &x)
{
  int dim = x.dimension();
  m_currentSet = m_x = x;
  m_r0 = VectorType(dim);
  split(m_x,m_r0);
  m_C = MatrixType::Identity(dim);
  m_B = MatrixType::Identity(dim);
  m_r = VectorType(dim);
  m_r.clear();

  m_currentMatrix = m_Bjac = m_Cjac = m_D = MatrixType::Identity(dim);
  m_R = MatrixType(dim,dim);
  m_R.clear();
  m_R0 = MatrixType(dim,dim);
  m_R0.clear();
  
  return *this;
}

template<typename MatrixType>
C1Rect2<MatrixType>& C1Rect2<MatrixType>::operator=(const C1Rect2 &S)
{
  if(&S==this)
    return *this;
  m_x = S.m_x;
  m_B = S.m_B;
  m_r = S.m_r;
  m_C = S.m_C;
  m_r0 = S.m_r0;
  
  m_D = S.m_D;
  m_Bjac = S.m_Bjac;
  m_R = S.m_R;
  m_Cjac = S.m_Cjac;
  m_R0 = S.m_R0;
  m_currentSet = S.m_currentSet;
  m_currentMatrix = S.m_currentMatrix;
  
  m_sizeFactor = S.m_sizeFactor;
  delete m_Norm;
  m_Norm = S.m_Norm->clone();
  return *this;
}

}} // namespace capd::dynset

#endif // _CAPD_DYNSET_C1RECT2_HPP_ 

/// @}
