/// @addtogroup dynsys
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file BasicTaylor.h
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

/* Author: Daniel Wilczak 2001-2005 */

#ifndef _CAPD_DYNSYS_BASICTAYLOR_H_ 
#define _CAPD_DYNSYS_BASICTAYLOR_H_ 

#include <string>

namespace capd{
namespace dynsys{
    /* MapT constraints:
     *  type definitions:
     *   - FunctionType
     *   - MatrixType
     *   - MatrixType::RowVectorType
     *   - MatrixType::ScalarType
     *
     *  methods:
     *   - getDimension         - dimension of the space
     *   - getOrder, setOrder   - order of Taylor expansion
     *   - operator()(a, i)     - computes coeffs a_i from a = a_{i-1}
     *   - variational(x_coeff, m_F, m_order);  
     *
     */

template<typename MapT>
class BasicTaylor
{
public:
  typedef MapT MapType;
  typedef typename MapT::FunctionType FunctionType;
  typedef typename MapType::MatrixType MatrixType;
  typedef typename MatrixType::RowVectorType VectorType;
  typedef typename MatrixType::ScalarType ScalarType;

  BasicTaylor(MapType& field, int order, const ScalarType& step);
  virtual ~BasicTaylor();

  /// Computes image of vector v after one time step
  VectorType operator()(const VectorType &) const;

  /// Computes image of vector v and derivatives of a flow with respect to init condition (v,identity)
  VectorType operator()(const VectorType &, MatrixType &derivative) const;

  const MapType& getField() const; ///< Returns vector field
  MapType& getField();
  
  int getOrder() const; ///< Returns Taylor method order
  virtual void setOrder(int newOrder); ///< Sets Taylor method order

  const ScalarType& getStep() const; ///< Returns current time step
  void setStep(const ScalarType& newStep); ///< Sets time step

  /// returns coefficients of Taylor expansion of the solution of ODE
  VectorType *getCoeff();

  /// returns coefficients of Taylor expansion of the monodromy matrix
  MatrixType *getMatrixCoeff();

protected:
  void operator=(const BasicTaylor&){} /// we do not allow copying of objects
  BasicTaylor(const BasicTaylor &){}    /// we do not allow copying of objects 

  void computeMatrixCoeff(int, const MatrixType& H) const;
  MapType* m_vField;

  int m_order;
  ScalarType m_step;
  VectorType *m_coeff;
  MatrixType *m_matrixCoeff;
  int m_dimension;

  MatrixType *m_F; // temp table used in the computations of variational part
};

// -----------------------------------------------------------------------------

template<typename MapType>
inline const MapType& BasicTaylor<MapType>::getField() const
{
  return *m_vField;
}

template<typename MapType>
inline MapType& BasicTaylor<MapType>::getField() 
{
  return *m_vField;
}

template<typename MapType>
inline int BasicTaylor<MapType>::getOrder() const
{
  return m_order;
}

template<typename MapType>
inline const typename BasicTaylor<MapType>::ScalarType& BasicTaylor<MapType>::getStep() const
{
  return m_step;
}

template<typename MapType>
inline void BasicTaylor<MapType>::setStep(const ScalarType& newStep)
{
  m_step = newStep;
}

template<typename MapType>
inline typename BasicTaylor<MapType>::VectorType* BasicTaylor<MapType>::getCoeff()
{
  return m_coeff;
}

template<typename MapType>
inline typename BasicTaylor<MapType>::MatrixType* BasicTaylor<MapType>::getMatrixCoeff()
{
  return m_matrixCoeff;
}

}} // namespace capd::dynsys

#endif // _CAPD_DYNSYS_BASICTAYLOR_H_ 

/// @}
