/// @addtogroup map
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file Map.hpp
///
/// @author Daniel Wilczak
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 2000-2005 by the CAPD Group.
//
// This file constitutes a part of the CAPD library, 
// distributed under the terms of the GNU General Public License.
// Consult  http://capd.wsb-nlu.edu.pl/ for details. 

/* Author: Daniel Wilczak 2001-2004 */

#ifndef _CAPD_MAP_MAP_HPP_ 
#define _CAPD_MAP_MAP_HPP_ 

#include <string>
#include <sstream>
#include <stdexcept>
#include <vector>
#include "capd/map/Function.hpp"
#include "capd/map/Node.hpp"
#include "capd/map/Parser.h"
#include "capd/map/Map.h"

namespace capd{
namespace map{

template<typename MatrixType>
void Map<MatrixType>::createFromText(const std::string &s, int Order)
{
  int dim1;
  std::string *var = Parser::splitVariables("var:",s,dim1);
  std::string *tab = Parser::splitVariables("fun:",s,dim);

  fun = new FunctionType[dim];
  der = new FunctionType[dim*dim];
  order = Order;
  std::string temp, variables;
  int i,j;
  variables = "var:";
  variables += var[0];
  for(j=1;j<dim1;++j)
    variables += "," + var[j];
  variables += ";fun:";

  for(i=0;i<dim;++i)
  {
    fun[i] = variables + tab[i] + ";";
    fun[i].setOrder(order);
  }

  delete []var;
  delete []tab;

  for(i=0;i<dim;++i)
    for(j=0;j<dim;++j)
    {
      der[i*dim+j]=fun[i][j];
      der[i*dim+j].setOrder(order);
    }
}

/*___________________________________________*/

template<typename MatrixType>
void Map<MatrixType>::createFromObject(const Map<MatrixType> &m)
{
  dim = m.dim;
  fun = new FunctionType[dim];
  der = new FunctionType[dim*dim];
  int i;
  for(i=0;i<dim;++i)
  {
    fun[i] = m.fun[i];
    der[i] = m.der[i];
  }
  for(i=dim;i<dim*dim;++i)
    der[i] = m.der[i];
  order = m.order;
}

/*___________________________________________*/

template<typename MatrixType>
Map<MatrixType>::Map(void)
{
  order=0;
  dim = 1;
  fun = new FunctionType[dim];
  der = new FunctionType[dim];
}

/*___________________________________________*/

template<typename MatrixType>
Map<MatrixType>::Map(const char *s, int order)
{
  try{
    createFromText(s,order);
  } catch(std::runtime_error &r)
  {
    std::string re = "exception in constructor Map::Map(const char *, int)\n";
    re += r.what();
    throw std::runtime_error(re);
  }
}

/*___________________________________________*/

template<typename MatrixType>
Map<MatrixType>::Map(const std::string &f, int order)
{
  try{
    createFromText(f,order);
  } catch(std::runtime_error &r)
  {
    std::string re = "exception in constructor Map::Map(const string&, int)\n";
    re += r.what();
    throw std::runtime_error(re);
  }
}

/*___________________________________________*/

template<typename MatrixType>
Map<MatrixType>::Map(const Map<MatrixType> &m)
{
  createFromObject(m);
}

/*_____________________________________________*/

template<typename MatrixType>
Map<MatrixType>& Map<MatrixType>::operator=(const char *s)
{
  delete []fun;
  delete []der;

  try{
    createFromText(s,0);
  } catch(std::runtime_error &r)
  {
    std::string re = "exception in Map &Map::operator=(const char *)\n";
    re += r.what();
    throw std::runtime_error(re);
  }
   return *this;
}

/*_______________________________________________*/

template<typename MatrixType>
Map<MatrixType>& Map<MatrixType>::operator=(const std::string &f)
{
  delete []fun;
  delete []der;
  try{
    createFromText(f,0);
  } catch(std::runtime_error &r)
  {
    std::string re = "exception in Map &Map::operator=(const string &f)\n";
    re += r.what();
    throw std::runtime_error(re);
  }
  return *this;
}

/*________________________________________________*/

template<typename MatrixType>
Map<MatrixType>& Map<MatrixType>::operator=(const Map<MatrixType> &m)
{
  delete []fun;
  delete []der;
  createFromObject(m);
  return *this;
}


/*________________________________________________*/

template<typename MatrixType>
MatrixType Map<MatrixType>::operator[](const VectorType &v) const
{
  MatrixType res(dim,dim);
  for(int i=0;i<dim;++i)
    for(int j=0;j<dim;++j)
      res[i][j] = der[i*dim+j](v);
  return res;
}


/*________________________________________________*/


template<typename MatrixType>
void Map<MatrixType>::setOrder(int the_new)
{
  order = the_new;
  int i;
  for(i=0;i<dim;++i)
  {
    fun[i].setOrder(order);
    der[i].setOrder(order);
  }
  for(i=dim;i<dim*dim;++i)
    der[i].setOrder(order);
}


/*________________________________________________*/

template<typename MatrixType>
void Map<MatrixType>::variational(VectorType iv[], MatrixType im[], int p) const
{
  if(p<0)
  {
    throw std::out_of_range ("exception in void Map::variational(vectorType iv[], matrixType im[], int p)\nnegative index of Taylor coefficient");
  }

  if(p>order)
  {
    std::ostringstream s;
    s << "exception in void Map::variational(vectorType iv[], matrixType im[], int p)\n";
    s << "Cannot compute " << p << "-th Taylor coefficient\n";
    s << "The order is " << order;
    throw std::out_of_range (s.str());
  }

  for(int i=0;i<dim;++i)
    for(int j=0;j<dim;++j)
      for(int r=0;r<=p;++r)
        im[r][i][j] = der[i*dim+j](iv[r],r);
}

/*________________________________________________*/

template<typename MatrixType>
void Map<MatrixType>::setParameter(const std::string &name, const ScalarType &value)
{
  for(int i=0;i<dim;++i)
  {
    fun[i].setParameter(name,value);
    for(int j=0;j<dim;++j)
      der[i*dim+j].setParameter(name,value);
  }
}

/*_______________________________________________*/

template<typename MatrixType>
typename Map<MatrixType>::VectorType Map<MatrixType>::operator()(const VectorType &v, int i) const
{
  VectorType result(dim);
  if (i<0)
    return result;
  for(int j=0;j<dim;++j)
    result[j] = fun[j](v,i);
  return result;
}

}} // namespace capd::map

#endif // _CAPD_MAP_MAP_HPP_ 

/// @}
